# -*- coding: utf-8 -*-
# Please don't import anything in this file to avoid issues when it is imported in setup.py

__version__ = "0.4.1"

CMD_NAME = "wireviz"  # Lower case command and module name
APP_NAME = "WireViz"  # Application name in texts meant to be human readable
APP_URL = "https://github.com/wireviz/WireViz"
