<!DOCTYPE html>
<html lang="en"><head>
  <meta charset="UTF-8">
  <meta name="generator" content="<!-- %generator% -->">
  <title><!-- %title% --></title>
  <style>

    #bom table, th, td {
      border: 1px solid black;
      border-collapse: collapse;
    }

    #bom th, td {
      padding: 4px;
      text-align: left;
    }

    .bom_col_qty {
      text-align: right;
    }

  </style>
</head><body style="font-family:<!-- %fontname% -->;background-color:<!-- %bgcolor% -->">
<h1><!-- %title% --></h1>
<h2>Diagram</h2>

<div id="description">
  <!-- %description% -->
</div>

<div id="diagram">
  <!-- %diagram% -->
</div>

<div id="notes">
  <!-- %notes% -->
</div>

<h2>Bill of Materials</h2>

<div id="bom">
  <!-- %bom% -->
</div>

</body></html>
